/* config.h.in.  Generated for macintosh builds.  */

/* Define if you have <sys/wait.h> that is POSIX.1 compatible.  */
/* #undef HAVE_SYS_WAIT_H */

/* Define to `int' if <sys/types.h> doesn't define.  */
/* #undef pid_t */

/* Define if you have the ANSI C header files.  */
#define STDC_HEADERS 1

/* Define if you want support for output to stdout */
/* #undef DRV_STDOUT */

/* Define if your system supports binary pipes (i.e. Unix) */
/* #undef DRV_PIPE */

/* Define if you want an .aiff file writer driver */
#define DRV_AIFF 1

/* Define if you want a raw pcm data file writer driver */
#define DRV_RAW 1

/* Define if you want a .wav file writer driver */
#define DRV_WAV 1

/* Define this if you want the Mac Sound Manager driver */
#define DRV_MAC 1

/* Define if you want a debug version of the library */
/* #undef MIKMOD_DEBUG */

/* Define if your system defines random(3) and srandom(3) in math.h instead
   of stdlib.h */
/* #undef SRANDOM_IN_MATH_H */

/* Define if you have the setenv function.  */
/* #undef HAVE_SETENV */

/* Define if you have the snprintf function.  */
#define HAVE_SNPRINTF 1

/* Define if you have the srandom function.  */
/* #undef HAVE_SRANDOM */

/* Define if you have the strcasecmp function.  */
#define HAVE_STRCASECMP 1

/* Define if you have the strdup function.  */
#define HAVE_STRDUP 1

/* Define if you have the strstr function.  */
#define HAVE_STRSTR 1

/* Define if you have the <fcntl.h> header file.  */
#define HAVE_FCNTL_H 1

/* Define if you have the <limits.h> header file. */
#define HAVE_LIMITS_H 1

/* Define if you have the <malloc.h> header file.  */
/* #undef HAVE_MALLOC_H */

/* Define if you have the <memory.h> header file.  */
/* #undef HAVE_MEMORY_H */

/* Define if you have the <strings.h> header file.  */
/* #undef HAVE_STRINGS_H */

/* Define if you have the <unistd.h> header file.  */
#define HAVE_UNISTD_H 1

